@@ -6,12 +6,17 @@ use 5.006;
 use strict;
 use warnings;
 
+use Log::Log4perl;
 use Log::Log4perl::Level;
 use Log::Log4perl::Layout;
 use Log::Log4perl::Appender;
+use Log::Log4perl::Appender::String;
 use Log::Log4perl::Filter;
 use Carp;
 
+$Carp::Internal{"Log::Log4perl"}++;
+$Carp::Internal{"Log::Log4perl::Logger"}++;
+
 use constant _INTERNAL_DEBUG => 0;
 
     # Initialization
@@ -21,8 +26,29 @@ our %APPENDER_BY_NAME = ();
 our $INITIALIZED = 0;
 our $NON_INIT_WARNED;
 
+    # Define the default appender that's used for formatting
+    # warn/die/croak etc. messages.
+our $STRING_APP_NAME = "_l4p_warn";
+our $STRING_APP      = Log::Log4perl::Appender->new(
+                          "Log::Log4perl::Appender::String",
+                          name => $STRING_APP_NAME);
+$STRING_APP->layout(Log::Log4perl::Layout::PatternLayout->new("%m"));
+our $STRING_APP_CODEREF = generate_coderef([[$STRING_APP_NAME, $STRING_APP]]);
+
 __PACKAGE__->reset();
 
+###########################################
+sub warning_render {
+###########################################
+    my($logger, @message) = @_;
+
+    $STRING_APP->string("");
+    $STRING_APP_CODEREF->($logger, 
+                          @message, 
+                          Log::Log4perl::Level::to_level($ALL));
+    return $STRING_APP->string();
+}
+
 ##################################################
 sub cleanup {
 ##################################################
@@ -42,6 +68,9 @@ sub cleanup {
     foreach my $appendername (keys %APPENDER_BY_NAME){
         if (exists $APPENDER_BY_NAME{$appendername} &&
             exists $APPENDER_BY_NAME{$appendername}->{appender}) {
+                # Destroy the specific appender
+            my $appref = $APPENDER_BY_NAME{$appendername}->{appender};
+            eval { $appref->DESTROY() };
                 # Destroy L4p::Appender
             $APPENDER_BY_NAME{$appendername}->DESTROY();
             delete $APPENDER_BY_NAME{$appendername}->{appender};
@@ -49,6 +78,8 @@ sub cleanup {
         delete $APPENDER_BY_NAME{$appendername};
     }
     %APPENDER_BY_NAME   = ();
+
+    undef $INITIALIZED;
 }
 
 ##################################################
@@ -329,14 +360,13 @@ sub generate_is_xxx_coderef {
 ##################################################
     my($return_token) = @_;
 
-    my $coderef    = '';
-    my $watch_code = '';
+    my $coderef    = sub { $return_token };
 
     if (defined $Log::Log4perl::Config::WATCHER) {
 
         my $cond = generate_watch_conditional();
 
-        $watch_code = <<EOL;
+        my $watch_code = <<EOL;
         my(\$logger, \$subname) = \@_;
         if($cond) {
             Log::Log4perl->init_and_watch();
@@ -344,13 +374,13 @@ sub generate_is_xxx_coderef {
             return \$logger->\$subname();
         }
 EOL
-    }
 
-    my $code = <<EOL;
-    \$coderef = sub { $watch_code return $return_token; };
+        my $code = <<EOL;
+        \$coderef = sub { $watch_code return $return_token; };
 EOL
 
-    eval $code or die "$@";
+        eval $code or die "$@";
+    }
 
     return $coderef;
 }
@@ -375,6 +405,12 @@ sub generate_watch_code {
             Log::Log4perl->init_and_watch();
                        
             my \$methodname = lc(\$level);
+
+                # Bump up the caller level by two, since
+                # we've artifically introduced additional levels.
+            local(\$Log::Log4perl::caller_depth);
+            \$Log::Log4perl::caller_depth += 2;
+
             \$logger->\$methodname(\@_); # send the message
                                          # to the new configuration
             return;        #and return, we're done with this incarnation
@@ -804,15 +840,14 @@ sub and_warn {
 #######################################################
   my $self = shift;
   my $msg = join("", @_[0 .. $#_]);
-  CORE::warn(callerline($msg));
+  CORE::warn(callerline($self->warning_render(@_[0 .. $#_])));
 }
 
 #######################################################
 sub and_die {
 #######################################################
   my $self = shift;
-  my $msg = join("", @_[0 .. $#_]);
-  die(callerline($msg));
+  die(callerline($self->warning_render(@_[0 .. $#_])));
 }
 
 ##################################################
@@ -842,16 +877,27 @@ sub logdie {
     $self->fatal(@chomped);
     $Log::Log4perl::caller_depth--;
   }
-  # no matter what, we die... 'cuz logdie wants you to die.
-  $self->and_die(@_);
+
+  $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
+      $self->and_die(@_) : 
+        exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
 ##################################################
-# for die and warn, carp long/shortmess return line #s and the like
+sub logexit {
 ##################################################
-sub noop {
-##################################################
-  return @_;
+  my $self = shift;
+
+  if ($self->is_fatal()) {
+        # Since we're one caller level off now, compensate for that.
+    $Log::Log4perl::caller_depth++;
+    my @chomped = @_;
+    chomp($chomped[-1]);
+    $self->fatal(@chomped);
+    $Log::Log4perl::caller_depth--;
+  }
+
+  exit $Log::Log4perl::LOGEXIT_CODE;
 }
 
 ##################################################
@@ -859,12 +905,18 @@ sub noop {
 sub logcluck {
 ##################################################
   my $self = shift;
+
+  local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+  my $msg = $self->warning_render(@_);
+
   if ($self->is_warn()) {
-    my $message = Carp::longmess(@_);
+    my $message = Carp::longmess($msg);
+    local $Log::Log4perl::caller_depth = 
+          $Log::Log4perl::caller_depth + 1;
     foreach (split(/\n/, $message)) {
       $self->warn("$_\n");
     }
-    CORE::warn(noop($message));
+    Carp::cluck($msg);
   }
 }
 
@@ -872,14 +924,20 @@ sub logcluck {
 sub logcarp {
 ##################################################
   my $self = shift;
+
+  local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+  my $msg = $self->warning_render(@_);
+
   if ($self->is_warn()) {
-    my $message = Carp::shortmess(@_);
+    my $message = Carp::shortmess($msg);
+    local $Log::Log4perl::caller_depth = 
+          $Log::Log4perl::caller_depth + 1;
     foreach (split(/\n/, $message)) {
       $self->warn("$_\n");
     }
-    CORE::warn(noop($message));
+    Carp::carp($msg) if $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR;
   }
-} 
+}
 
 ##################################################
 # croaks and confess are FATAL level
@@ -887,28 +945,44 @@ sub logcarp {
 sub logcroak {
 ##################################################
   my $self = shift;
-  my $message = Carp::shortmess(@_);
+
+  local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+  my $msg = $self->warning_render(@_);
+
+  my $message = Carp::shortmess($msg);
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
   if ($self->is_fatal()) {
     foreach (split(/\n/, $message)) {
       $self->fatal("$_\n");
     }
   }
-  # again, we die no matter what
-  die(noop($message));
+
+  $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
+      Carp::croak($msg) : 
+        exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
 ##################################################
 sub logconfess {
 ##################################################
   my $self = shift;
-  my $message = Carp::longmess(@_);
+
+  my $msg = $self->warning_render(@_);
+  local $Carp::CarpLevel = 2;
+
+  my $message = Carp::longmess($msg);
+  local $Log::Log4perl::caller_depth = 
+        $Log::Log4perl::caller_depth + 1;
   if ($self->is_fatal()) {
     foreach (split(/\n/, $message)) {
       $self->fatal("$_\n");
     }
   }
-  # again, we die no matter what
-  die(noop($message));
+
+  $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
+      confess($msg) :
+        exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
 ##################################################
@@ -927,12 +1001,18 @@ sub error_warn {
 sub error_die {
 ##################################################
   my $self = shift;
+
+  my $msg = $self->warning_render(@_);
+
   if ($self->is_error()) {
     $Log::Log4perl::caller_depth++;
-    $self->error(@_);
+    $self->error($msg);
     $Log::Log4perl::caller_depth--;
   }
-  $self->and_die(@_);
+
+  $Log::Log4perl::LOGDIE_MESSAGE_ON_STDERR ? 
+      $self->and_die($msg) :
+        exit($Log::Log4perl::LOGEXIT_CODE);
 }
 
 ##################################################
